@@ -21,7 +21,7 @@ use strict;
 use vars qw($VERSION);
 use Image::ExifTool qw(:DataAccess :Utils);
 
-$VERSION = '1.06';
+$VERSION = '1.13';
 
 sub ProcessPEResources($$);
 sub ProcessPEVersion($$);
@@ -415,7 +415,7 @@ my %languageCode = (
     InternalName    => { },
     LegalCopyright  => { },
     LegalTrademarks => { },
-    OriginalFilename=> { },
+    OriginalFilename=> { Name => 'OriginalFileName' },
     PrivateBuild    => { },
     ProductName     => { },
     ProductVersion  => { },
@@ -598,8 +598,40 @@ my %languageCode = (
             7 => 'Dynamic link editor',
             8 => 'Dynamically bound bundle',
             9 => 'Shared library stub for static linking',
+            # (the following from Apple loader.h header file)
+            10 => 'Debug information',
+            11 => 'x86_64 kexts',
         },
     },
+    6 => {
+        Name => 'ObjectFlags',
+        PrintHex => 1,
+        # ref Apple loader.h header file
+        PrintConv => { BITMASK => {
+             0 => 'No undefs',
+             1 => 'Incrementa link',
+             2 => 'Dyld link',
+             3 => 'Bind at load',
+             4 => 'Prebound',
+             5 => 'Split segs',
+             6 => 'Lazy init',
+             7 => 'Two level',
+             8 => 'Force flat',
+             9 => 'No multi defs',
+             10 => 'No fix prebinding',
+             11 => 'Prebindable',
+             12 => 'All mods bound',
+             13 => 'Subsections via symbols',
+             14 => 'Canonical',
+             15 => 'Weak defines',
+             16 => 'Binds to weak',
+             17 => 'Allow stack execution',
+             18 => 'Dead strippable dylib',
+             19 => 'Root safe',
+             20 => 'No reexported dylibs',
+             21 => 'Random address',
+        }},
+    },
 );
 
 # Information extracted from PEF (Classic MacOS executable) file header
@@ -702,6 +734,29 @@ my %languageCode = (
     },
 );
 
+# Information extracted from static library archives
+# (ref http://opensource.apple.com//source/xnu/xnu-1456.1.26/EXTERNAL_HEADERS/ar.h)
+%Image::ExifTool::EXE::AR = (
+    PROCESS_PROC => \&Image::ExifTool::ProcessBinaryData,
+    GROUPS => { 2 => 'Other' },
+    NOTES => q{
+        Information extracted from static libraries.
+    },
+  #  0  string[16] ar_name
+    16 => {
+        Name => 'CreateDate',
+        Groups => { 2 => 'Time' },
+        Format => 'string[12]',
+        ValueConv => 'ConvertUnixTime($val,1)',
+        PrintConv => '$self->ConvertDateTime($val)',
+    },
+  # 28  string[6]  ar_uid
+  # 34  string[6]  ar_gid
+  # 40  string[8]  ar_mode
+  # 48  string[10] ar_size
+  # 58  string[2]  terminator "`\n"
+);
+
 # Microsoft compiled help format (ref http://www.russotto.net/chm/chmformat.html)
 %Image::ExifTool::EXE::CHM = (
     PROCESS_PROC => \&Image::ExifTool::ProcessBinaryData,
@@ -726,37 +781,37 @@ my %languageCode = (
 # Returns: 1 on success, 0 if this wasn't a valid CHM file
 sub ProcessCHM($$)
 {
-    my ($exifTool, $dirInfo) = @_;
+    my ($et, $dirInfo) = @_;
     my $raf = $$dirInfo{RAF};
     my $buff;
 
     return 0 unless $raf->Read($buff, 56) == 56 and
         $buff =~ /^ITSF.{20}\x10\xfd\x01\x7c\xaa\x7b\xd0\x11\x9e\x0c\0\xa0\xc9\x22\xe6\xec/s;
     my $tagTablePtr = GetTagTable('Image::ExifTool::EXE::CHM');
-    $exifTool->SetFileType();
+    $et->SetFileType();
     SetByteOrder('II');
-    $exifTool->ProcessDirectory({ DataPt => \$buff }, $tagTablePtr);
+    $et->ProcessDirectory({ DataPt => \$buff }, $tagTablePtr);
     return 1;
 }
 
 #------------------------------------------------------------------------------
 # Read Unicode string (null terminated) from resource
-# Inputs: 0) data ref, 1) start offset, 2) optional ExifTool object ref
+# Inputs: 0) data ref, 1) start offset, 2) data end, 3) optional ExifTool object ref
 # Returns: 0) Unicode string translated to UTF8, or current CharSet with ExifTool ref
 #          1) end pos (rounded up to nearest 4 bytes)
-sub ReadUnicodeStr($$;$)
+sub ReadUnicodeStr($$$;$)
 {
-    my ($dataPt, $pos, $exifTool) = @_;
-    my $len = length $$dataPt;
+    my ($dataPt, $pos, $end, $et) = @_;
+    $end = length $$dataPt if $end > length $$dataPt;   # (be safe)
     my $str = '';
-    while ($pos + 2 <= $len) {
+    while ($pos + 2 <= $end) {
         my $ch = substr($$dataPt, $pos, 2);
         $pos += 2;
         last if $ch eq "\0\0";
         $str .= $ch;
     }
     $pos += 2 if $pos & 0x03;
-    my $to = $exifTool ? $exifTool->Options('Charset') : 'UTF8';
+    my $to = $et ? $et->Options('Charset') : 'UTF8';
     return (Image::ExifTool::Decode(undef,$str,'UCS2','II',$to), $pos);
 }
 
@@ -766,7 +821,7 @@ sub ReadUnicodeStr($$;$)
 # Returns: true on success
 sub ProcessPEVersion($$)
 {
-    my ($exifTool, $dirInfo) = @_;
+    my ($et, $dirInfo) = @_;
     my $dataPt = $$dirInfo{DataPt};
     my $pos = $$dirInfo{DirStart};
     my $end = $pos + $$dirInfo{DirLen};
@@ -780,10 +835,10 @@ sub ProcessPEVersion($$)
         $valLen = Get16u($dataPt, $pos + 2);
         $type = Get16u($dataPt, $pos + 4);
         return 0 unless $len or $valLen;  # prevent possible infinite loop
-        ($string, $strEnd) = ReadUnicodeStr($dataPt, $pos + 6);
+        ($string, $strEnd) = ReadUnicodeStr($dataPt, $pos + 6, $pos + $len);
         return 0 if $strEnd + $valLen > $end;
         unless ($index or $string eq 'VS_VERSION_INFO') {
-            $exifTool->Warn('Invalid Version Info block');
+            $et->Warn('Invalid Version Info block');
             return 0;
         }
         if ($string eq 'VS_VERSION_INFO') {
@@ -791,7 +846,7 @@ sub ProcessPEVersion($$)
             $$dirInfo{DirStart} = $strEnd;
             $$dirInfo{DirLen} = $valLen;
             my $subTablePtr = GetTagTable('Image::ExifTool::EXE::PEVersion');
-            $exifTool->ProcessDirectory($dirInfo, $subTablePtr);
+            $et->ProcessDirectory($dirInfo, $subTablePtr);
             $pos = $strEnd + $valLen;
         } elsif ($string eq 'StringFileInfo' and $valLen == 0) {
             $pos += $len;
@@ -802,8 +857,9 @@ sub ProcessPEVersion($$)
                 $len = Get16u($dataPt, $pt);
                 $valLen = Get16u($dataPt, $pt + 2);
                 # $type = Get16u($dataPt, $pt + 4);
+                my $entryEnd = $pt + $len;
                 # get tag ID (converted to UTF8)
-                ($string, $pt) = ReadUnicodeStr($dataPt, $pt + 6);
+                ($string, $pt) = ReadUnicodeStr($dataPt, $pt + 6, $entryEnd);
                 unless ($index) {
                     # separate the language code and character set
                     # (not sure what the CharacterSet tag is for, but the string
@@ -814,8 +870,8 @@ sub ProcessPEVersion($$)
                         $char = substr($string, 4);
                         $string = substr($string, 0, 4);
                     }
-                    $exifTool->HandleTag($tagTablePtr, 'LanguageCode', uc $string);
-                    $exifTool->HandleTag($tagTablePtr, 'CharacterSet', uc $char) if $char;
+                    $et->HandleTag($tagTablePtr, 'LanguageCode', uc $string);
+                    $et->HandleTag($tagTablePtr, 'CharacterSet', uc $char) if $char;
                     next;
                 }
                 my $tag = $string;
@@ -828,11 +884,13 @@ sub ProcessPEVersion($$)
                 }
                 # get tag value (converted to current Charset)
                 if ($valLen) {
-                    ($string, $pt) = ReadUnicodeStr($dataPt, $pt, $exifTool);
+                    ($string, $pt) = ReadUnicodeStr($dataPt, $pt, $entryEnd, $et);
                 } else {
                     $string = '';
                 }
-                $exifTool->HandleTag($tagTablePtr, $tag, $string);
+                $et->HandleTag($tagTablePtr, $tag, $string);
+                # step to next entry (padded to an even word)
+                $pt = ($entryEnd + 3) & 0xfffffffc;
             }
         } else {
             $pos += $len + $valLen;
@@ -848,12 +906,12 @@ sub ProcessPEVersion($$)
 # Returns: true on success
 sub ProcessPEResources($$)
 {
-    my ($exifTool, $dirInfo) = @_;
+    my ($et, $dirInfo) = @_;
     my $raf = $$dirInfo{RAF};
     my $base = $$dirInfo{Base};
     my $dirStart = $$dirInfo{DirStart} + $base;
     my $level = $$dirInfo{Level} || 0;
-    my $verbose = $exifTool->Options('Verbose');
+    my $verbose = $et->Options('Verbose');
     my ($buff, $buf2, $item);
 
     return 0 if $level > 10;    # protect against deep recursion
@@ -873,7 +931,7 @@ sub ProcessPEResources($$)
             my $resType = $resourceType{$name} || sprintf('Unknown (0x%x)', $name);
             # ignore everything but the Version resource unless verbose
             if ($verbose) {
-                $exifTool->VPrint(0, "$resType resource:\n");
+                $et->VPrint(0, "$resType resource:\n");
             } else {
                 next unless $resType eq 'Version';
             }
@@ -883,7 +941,7 @@ sub ProcessPEResources($$)
             # descend into next directory level
             $$dirInfo{DirStart} = $entryPos & 0x7fffffff;
             $$dirInfo{Level} = $level + 1;
-            ProcessPEResources($exifTool, $dirInfo) or return 0;
+            ProcessPEResources($et, $dirInfo) or return 0;
             --$$dirInfo{Level};
         } elsif ($$dirInfo{ResType} eq 'Version' and $level == 2 and
             not $$dirInfo{GotVersion}) # (only process first Version resource)
@@ -903,12 +961,12 @@ sub ProcessPEResources($$)
             }
             return 0 unless $filePos;
             $raf->Seek($filePos, 0) and $raf->Read($buf2, $len) == $len or return 0;
-            ProcessPEVersion($exifTool, {
+            ProcessPEVersion($et, {
                 DataPt   => \$buf2,
                 DataLen  => $len,
                 DirStart => 0,
                 DirLen   => $len,
-            }) or $exifTool->Warn('Possibly corrupt Version resource');
+            }) or $et->Warn('Possibly corrupt Version resource');
             $$dirInfo{GotVersion} = 1;  # set flag so we don't do this again
         }
     }
@@ -921,7 +979,7 @@ sub ProcessPEResources($$)
 # Returns: true on success
 sub ProcessPEDict($$)
 {
-    my ($exifTool, $dirInfo) = @_;
+    my ($et, $dirInfo) = @_;
     my $raf = $$dirInfo{RAF};
     my $dataPt = $$dirInfo{DataPt};
     my $dirLen = length($$dataPt);
@@ -945,21 +1003,90 @@ sub ProcessPEDict($$)
         ) if $name eq ".rsrc\0\0\0" and not %dirInfo;
     }
     # process the first resource section
-    ProcessPEResources($exifTool, \%dirInfo) or return 0 if %dirInfo;
+    ProcessPEResources($et, \%dirInfo) or return 0 if %dirInfo;
     return 1;
 }
 
+#------------------------------------------------------------------------------
+# Override file type if necessary for Mach object files and libraries
+# Inputs: 0) ExifTool ref, 1) ObjectFileType number, 2) flag for fat binary
+my %machOverride = (
+    1 => [ 'object file', 'O' ],
+    6 => [ 'dynamic link library', 'DYLIB' ],
+    8 => [ 'dynamic bound bundle', 'DYLIB' ],
+    9 => [ 'dynamic link library stub', 'DYLIB' ],
+);
+sub MachOverride($$;$)
+{
+    my ($et, $objType, $fat) = @_;
+    my $override = $machOverride{$objType};
+    if ($override) {
+        my $desc = 'Mach-O ' . ($fat ? 'fat ' : '') . $$override[0];
+        $et->OverrideFileType($desc, undef, $$override[1]);
+    }
+}
+
+#------------------------------------------------------------------------------
+# Extract tags from Mach header
+# Inputs: 0) ExifTool ref, 1) data ref, 2) flag to extract object type
+# Returns: true if Mach header was found
+# Mach type based on magic number
+# [bit depth, byte order starting with "Little" or "Big"]
+my %machType = (
+    "\xfe\xed\xfa\xce" => ['32 bit', 'Big endian'],
+    "\xce\xfa\xed\xfe" => ['32 bit', 'Little endian'],
+    "\xfe\xed\xfa\xcf" => ['64 bit', 'Big endian'],
+    "\xcf\xfa\xed\xfe" => ['64 bit', 'Little endian'],
+);
+sub ExtractMachTags($$;$)
+{
+    my ($et, $dataPt, $doObj) = @_;
+    # get information about mach header based on the magic number (first 4 bytes)
+    my $info = $machType{substr($$dataPt, 0, 4)};
+    if ($info) {
+        # Mach header structure:
+        #  0 int32u magic
+        #  4 int32u cputype
+        #  8 int32u cpusubtype
+        # 12 int32u filetype
+        # 16 int32u ncmds
+        # 20 int32u sizeofcmds
+        # 24 int32u flags
+        my $tagTablePtr = GetTagTable('Image::ExifTool::EXE::MachO');
+        SetByteOrder($$info[1]);
+        my $cpuType = Get32s($dataPt, 4);
+        my $subType = Get32s($dataPt, 8);
+        $et->HandleTag($tagTablePtr, 0, $$info[0]);
+        $et->HandleTag($tagTablePtr, 1, $$info[1]);
+        $et->HandleTag($tagTablePtr, 3, $cpuType);
+        $et->HandleTag($tagTablePtr, 4, "$cpuType $subType");
+        if ($doObj) {
+            my $objType = Get32u($dataPt, 12);
+            my $flags = Get32u($dataPt, 24);
+            $et->HandleTag($tagTablePtr, 5, $objType);
+            $et->HandleTag($tagTablePtr, 6, $flags);
+            # override file type if this is an object file or library
+            MachOverride($et, $objType);
+        } else { # otherwise this was a static library
+            $et->OverrideFileType('Mach-O static library', undef, 'A');
+        }
+        return 1;
+    }
+    return 0;
+}
+
 #------------------------------------------------------------------------------
 # Extract information from an EXE file
 # Inputs: 0) ExifTool object reference, 1) dirInfo reference
 # Returns: 1 on success, 0 if this wasn't a valid EXE file
 sub ProcessEXE($$)
 {
-    my ($exifTool, $dirInfo) = @_;
+    my ($et, $dirInfo) = @_;
     my $raf = $$dirInfo{RAF};
-    my ($buff, $buf2, $type, $tagTablePtr, %dirInfo);
+    my ($buff, $buf2, $type, $mime, $ext, $tagTablePtr, %dirInfo);
 
     my $size = $raf->Read($buff, 0x40) or return 0;
+    my $fast3 = $$et{OPTIONS}{FastScan} && $$et{OPTIONS}{FastScan} == 3;
 #
 # DOS and Windows EXE
 #
@@ -967,25 +1094,25 @@ sub ProcessEXE($$)
         # DOS/Windows executable
         # validate DOS header
         # (ref http://www.delphidabbler.com/articles?article=8&part=2)
-        #   0  magic   : int16u    # Magic number ("MZ")
-        #   2  cblp    : int16u    # Bytes on last page of file
-        #   4  cp      : int16u    # Pages in file
-        #   6  crlc    : int16u    # Relocations
-        #   8  cparhdr : int16u    # Size of header in paragraphs
-        #  10  minalloc: int16u    # Minimum extra paragraphs needed
-        #  12  maxalloc: int16u    # Maximum extra paragraphs needed
-        #  14  ss      : int16u    # Initial (relative) SS value
-        #  16  sp      : int16u    # Initial SP value
-        #  18  csum    : int16u    # Checksum
-        #  20  ip      : int16u    # Initial IP value
-        #  22  cs      : int16u    # Initial (relative) CS value
-        #  24  lfarlc  : int16u    # Address of relocation table
-        #  26  ovno    : int16u    # Overlay number
-        #  28  res     : int16u[4] # Reserved words
-        #  36  oemid   : int16u    # OEM identifier (for oeminfo)
-        #  38  oeminfo : int16u    # OEM info; oemid specific
-        #  40  res2    : int16u[10]# Reserved words
-        #  60  lfanew  : int32u;   # File address of new exe header
+        #   0 int16u     magic    - Magic number ("MZ")
+        #   2 int16u     cblp     - Bytes on last page of file
+        #   4 int16u     cp       - Pages in file
+        #   6 int16u     crlc     - Relocations
+        #   8 int16u     cparhdr  - Size of header in paragraphs
+        #  10 int16u     minalloc - Minimum extra paragraphs needed
+        #  12 int16u     maxalloc - Maximum extra paragraphs needed
+        #  14 int16u     ss       - Initial (relative) SS value
+        #  16 int16u     sp       - Initial SP value
+        #  18 int16u     csum     - Checksum
+        #  20 int16u     ip       - Initial IP value
+        #  22 int16u     cs       - Initial (relative) CS value
+        #  24 int16u     lfarlc   - Address of relocation table
+        #  26 int16u     ovno     - Overlay number
+        #  28 int16u[4]  res      - Reserved words
+        #  36 int16u     oemid    - OEM identifier (for oeminfo)
+        #  38 int16u     oeminfo  - OEM info; oemid specific
+        #  40 int16u[10] res2     - Reserved words
+        #  60 int32u;    lfanew   - File address of new exe header
         SetByteOrder('II');
         my ($cblp, $cp, $lfarlc, $lfanew) = unpack('x2v2x18vx34V', $buff);
         my $fileSize = ($cp - ($cblp ? 1 : 0)) * 512 + $cblp;
@@ -999,7 +1126,8 @@ sub ProcessEXE($$)
                 if ($size >= 0x40) { # NE header is 64 bytes (ref 2)
                     # check for DLL
                     my $appFlags = Get16u(\$buff, 0x0c);
-                    $type = 'Win16 ' . ($appFlags & 0x80 ? 'DLL' : 'EXE');
+                    $ext = $appFlags & 0x80 ? 'DLL' : 'EXE';
+                    $type = "Win16 $ext";
                     # offset 0x02 is 2 bytes with linker version and revision numbers
                     # offset 0x36 is executable type (2 = Windows)
                 }
@@ -1016,7 +1144,9 @@ sub ProcessEXE($$)
                     my $machine = $Image::ExifTool::EXE::Main{0}{PrintConv}{Get16u(\$buff, 4)} || '';
                     my $winType = $machine =~ /64/ ? 'Win64' : 'Win32';
                     my $flags = Get16u(\$buff, 22);
-                    $exifTool->SetFileType($winType . ' ' . ($flags & 0x2000 ? 'DLL' : 'EXE'));
+                    $ext = $flags & 0x2000 ? 'DLL' : 'EXE';
+                    $et->SetFileType("$winType $ext", undef, $ext);
+                    return 1 if $fast3;
                     # read the rest of the optional header if necessary
                     my $optSize = Get16u(\$buff, 20);
                     my $more = $optSize + 24 - $size;
@@ -1027,11 +1157,11 @@ sub ProcessEXE($$)
                             my $magic = Get16u(\$buff, 24);
                             # verify PE32/PE32+ magic number
                             unless ($magic == 0x10b or $magic == 0x20b) {
-                                $exifTool->Warn('Unknown PE magic number');
+                                $et->Warn('Unknown PE magic number');
                                 return 1;
                             }
                         } else {
-                            $exifTool->Warn('Error reading optional header');
+                            $et->Warn('Error reading optional header');
                         }
                     }
                     # process PE COFF file header
@@ -1043,7 +1173,7 @@ sub ProcessEXE($$)
                         DirStart => 4,
                         DirLen => $size,
                     );
-                    $exifTool->ProcessDirectory(\%dirInfo, $tagTablePtr);
+                    $et->ProcessDirectory(\%dirInfo, $tagTablePtr);
                     # process data dictionary
                     my $num = Get16u(\$buff, 6);    # NumberOfSections
                     if ($raf->Read($buff, 40 * $num) == 40 * $num) {
@@ -1051,15 +1181,17 @@ sub ProcessEXE($$)
                             RAF => $raf,
                             DataPt => \$buff,
                         );
-                        ProcessPEDict($exifTool, \%dirInfo) or $exifTool->Warn('Error processing PE data dictionary');
+                        ProcessPEDict($et, \%dirInfo) or $et->Warn('Error processing PE data dictionary');
                     }
                     return 1;
                 }
             } else {
                 $type = 'Virtual Device Driver';
+                $ext = '386';
             }
         } else {
             $type = 'DOS EXE';
+            $ext = 'exe';
         }
 #
 # Mach-O (Mac OS X)
@@ -1070,59 +1202,50 @@ sub ProcessEXE($$)
         $tagTablePtr = GetTagTable('Image::ExifTool::EXE::MachO');
         if ($1 eq "\xca\xfe\xba\xbe") {
             SetByteOrder('MM');
-            $exifTool->SetFileType('Mach-O fat binary executable');
+            $et->SetFileType('Mach-O fat binary executable', undef, '');
+            return 1 if $fast3;
             my $count = Get32u(\$buff, 4);  # get architecture count
             my $more = $count * 20 - ($size - 8);
             if ($more > 0) {
                 unless ($raf->Read($buf2, $more) == $more) {
-                    $exifTool->Warn('Error reading fat-arch headers');
+                    $et->Warn('Error reading fat-arch headers');
                     return 1;
                 }
                 $buff .= $buf2;
                 $size += $more;
             }
-            $exifTool->HandleTag($tagTablePtr, 2, $count);
+            $et->HandleTag($tagTablePtr, 2, $count);
             my $i;
             for ($i=0; $i<$count; ++$i) {
                 my $cpuType = Get32s(\$buff, 8 + $i * 20);
-                my $cpuSubtype = Get32u(\$buff, 12 + $i * 20);
-                $exifTool->HandleTag($tagTablePtr, 3, $cpuType);
-                $exifTool->HandleTag($tagTablePtr, 4, "$cpuType $cpuSubtype");
+                my $subType = Get32s(\$buff, 12 + $i * 20);
+                $et->HandleTag($tagTablePtr, 3, $cpuType);
+                $et->HandleTag($tagTablePtr, 4, "$cpuType $subType");
             }
             # load first Mach-O header to get the object file type
             my $offset = Get32u(\$buff, 16);
             if ($raf->Seek($offset, 0) and $raf->Read($buf2, 16) == 16) {
                 if ($buf2 =~ /^(\xfe\xed\xfa(\xce|\xcf)|(\xce|\xcf)\xfa\xed\xfe)/) {
                     SetByteOrder($buf2 =~ /^\xfe\xed/ ? 'MM' : 'II');
-                    my $objType = Get32s(\$buf2, 12);
-                    $exifTool->HandleTag($tagTablePtr, 5, $objType);
+                    my $objType = Get32u(\$buf2, 12);
+                    $et->HandleTag($tagTablePtr, 5, $objType);
+                    # override file type if this is a library or object file
+                    MachOverride($et, $objType, 'fat');
                 } elsif ($buf2 =~ /^!<arch>\x0a/) {
                     # .a libraries use this magic number
-                    $exifTool->HandleTag($tagTablePtr, 5, -1);
+                    $et->HandleTag($tagTablePtr, 5, -1);
+                    # override file type since this is a library
+                    $et->OverrideFileType('Mach-O fat static library', undef, 'A');
                 } else {
-                    $exifTool->Warn('Unrecognized object file type');
+                    $et->Warn('Unrecognized object file type');
                 }
             } else {
-                $exifTool->Warn('Error reading file');
+                $et->Warn('Error reading file');
             }
-       } elsif ($size >= 16) {
-            $exifTool->SetFileType('Mach-O executable');
-            my $info = {
-                "\xfe\xed\xfa\xce" => ['32 bit', 'Big endian'],
-                "\xce\xfa\xed\xfe" => ['32 bit', 'Little endian'],
-                "\xfe\xed\xfa\xcf" => ['64 bit', 'Big endian'],
-                "\xcf\xfa\xed\xfe" => ['64 bit', 'Little endian'],
-            }->{substr($buff, 0, 4)};
-            my $byteOrder = ($buff =~ /^\xfe/) ? 'MM' : 'II';
-            SetByteOrder($byteOrder);
-            my $cpuType = Get32s(\$buff, 4);
-            my $cpuSubtype = Get32s(\$buff, 8);
-            my $objType = Get32s(\$buff, 12);
-            $exifTool->HandleTag($tagTablePtr, 0, $$info[0]);
-            $exifTool->HandleTag($tagTablePtr, 1, $$info[1]);
-            $exifTool->HandleTag($tagTablePtr, 3, $cpuType);
-            $exifTool->HandleTag($tagTablePtr, 4, "$cpuType $cpuSubtype");
-            $exifTool->HandleTag($tagTablePtr, 5, $objType);
+        } elsif ($size >= 16) {
+            $et->SetFileType('Mach-O executable', undef, '');
+            return 1 if $fast3;
+            ExtractMachTags($et, \$buff, 1);
         }
         return 1;
 #
@@ -1130,7 +1253,8 @@ sub ProcessEXE($$)
 #
     } elsif ($buff =~ /^Joy!peff/ and $size > 12) {
         # ref http://developer.apple.com/documentation/mac/pdf/MacOS_RT_Architectures.pdf
-        $exifTool->SetFileType('Classic MacOS executable');
+        $et->SetFileType('Classic MacOS executable', undef, '');
+        return 1 if $fast3;
         SetByteOrder('MM');
         $tagTablePtr = GetTagTable('Image::ExifTool::EXE::PEF');
         %dirInfo = (
@@ -1140,37 +1264,86 @@ sub ProcessEXE($$)
             DirStart => 0,
             DirLen => $size,
         );
-        $exifTool->ProcessDirectory(\%dirInfo, $tagTablePtr);
+        $et->ProcessDirectory(\%dirInfo, $tagTablePtr);
         return 1;
 #
 # ELF (Unix)
 #
     } elsif ($buff =~ /^\x7fELF/ and $size >= 16) {
-        $exifTool->SetFileType("ELF executable");
+        $et->SetFileType('ELF executable', undef, '');
+        return 1 if $fast3;
         SetByteOrder(Get8u(\$buff,5) == 1 ? 'II' : 'MM');
         $tagTablePtr = GetTagTable('Image::ExifTool::EXE::ELF');
         %dirInfo = (
-            DataPt => \$buff,
+            DataPt  => \$buff,
             DataPos => 0,
             DataLen => $size,
-            DirStart => 0,
-            DirLen => $size,
+            DirLen  => $size,
         );
-        $exifTool->ProcessDirectory(\%dirInfo, $tagTablePtr);
+        $et->ProcessDirectory(\%dirInfo, $tagTablePtr);
+        # override file type if this is a library or object file
+        my $override = {
+            1 => [ 'ELF object file', 'O' ],
+            3 => [ 'ELF shared library', 'SO' ],
+        }->{$$et{VALUE}{ObjectFileType} || 0};
+        $et->OverrideFileType($$override[0], undef, $$override[1]) if $override;
+        return 1;
+#
+# .a libraries
+#
+    } elsif ($buff =~ /^!<arch>\x0a/) {
+        $et->SetFileType('Static library', undef, 'A');
+        return 1 if $fast3;
+        my $pos = 8;    # current file position
+        my $max = 10;   # maximum number of archive files to check
+        # read into list of ar structures (each 60 bytes long):
+        while ($max-- > 0) {
+            # seek to start of the ar structure and read it
+            $raf->Seek($pos, 0) and $raf->Read($buff, 60) == 60 or last;
+            substr($buff, 58, 2) eq "`\n" or $et->Warn('Invalid archive header'), last;
+            unless ($tagTablePtr) {
+                # extract some information from first file in archive
+                $tagTablePtr = GetTagTable('Image::ExifTool::EXE::AR');
+                %dirInfo = (
+                    DataPt  => \$buff,
+                    DataPos => $pos,
+                );
+                $et->ProcessDirectory(\%dirInfo, $tagTablePtr);
+            }
+            my $name = substr($buff, 0, 16);
+            if ($name =~ m{^#1/(\d+) *$}) { # check for extended archive (BSD variant)
+                my $len = $1;
+                $len > 256 and $et->Warn('Invalid extended archive name length'), last;
+                # (we read the name here just to move the file pointer)
+                $raf->Read($name, $len) == $len or $et->Warn('Error reading archive name'), last;
+            }
+            my $arSize = substr($buff, 48, 10);
+            $arSize =~ s/^(\d+).*/$1/s or last;     # make sure archive size is a number
+            $raf->Read($buff, 28) == 28 or last;    # read (possible) Mach header
+            ExtractMachTags($et, \$buff) and last;  # try to extract tags
+            $pos += 60 + $arSize;   # step to next entry
+            ++$pos if $pos & 0x01;  # padded to an even byte
+        }
         return 1;
 #
 # various scripts (perl, sh, etc...)
 #
     } elsif ($buff =~ m{^#!\s*/\S*bin/(\w+)}) {
-        $type = "$1 script";
-#
-# .a libraries
-#
-    } elsif ($buff =~ /^!<arch>\x0a/) {
-        $type = 'Static library',
+        my $prog = $1;
+        $prog = $1 if $prog eq 'env' and $buff =~ /\b(perl|python|ruby|php)\b/;
+        $type = "$prog script";
+        $mime = "text/x-$prog";
+        $ext = {
+            perl   => 'pl',
+            python => 'py',
+            ruby   => 'rb',
+            php    => 'php',
+        }->{$1};
+        # use '.sh' for extension of all shell scripts
+        $ext = $prog =~ /sh$/ ? 'sh' : '' unless defined $ext;
     }
     return 0 unless $type;
-    $exifTool->SetFileType($type);
+    $et->SetFileType($type, $mime, $ext);
     return 1;
 }
 
@@ -1194,7 +1367,7 @@ library files.
 
 =head1 AUTHOR
 
-Copyright 2003-2013, Phil Harvey (phil at owl.phy.queensu.ca)
+Copyright 2003-2017, Phil Harvey (phil at owl.phy.queensu.ca)
 
 This library is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
